@@ -1,9 +1,9 @@
 /* Test of opening a file stream.
-   Copyright (C) 2007-2022 Free Software Foundation, Inc.
+   Copyright (C) 2007-2025 Free Software Foundation, Inc.
 
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
-   the Free Software Foundation; either version 3 of the License, or
+   the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.
 
    This program is distributed in the hope that it will be useful,
@@ -30,5 +30,6 @@ SIGNATURE_CHECK (fopen, FILE *, (char const *, char const *));
 int
 main (void)
 {
-  return test_fopen ();
+  int result = test_fopen ();
+  return (result ? result : test_exit_status);
 }
