@@ -1,5 +1,5 @@
 /* Python format strings.
-   Copyright (C) 2001-2004, 2006-2009 Free Software Foundation, Inc.
+   Copyright (C) 2001-2025 Free Software Foundation, Inc.
    Written by Bruno Haible <haible@clisp.cons.org>, 2001.
 
    This program is free software: you can redistribute it and/or modify
@@ -92,17 +92,10 @@ struct spec
   unsigned int directives;
   unsigned int named_arg_count;
   unsigned int unnamed_arg_count;
-  unsigned int allocated;
   struct named_arg *named;
   struct unnamed_arg *unnamed;
 };
 
-/* Locale independent test for a decimal digit.
-   Argument can be  'char' or 'unsigned char'.  (Whereas the argument of
-   <ctype.h> isdigit must be an 'unsigned char'.)  */
-#undef isdigit
-#define isdigit(c) ((unsigned int) ((c) - '0') < 10)
-
 
 static int
 named_arg_compare (const void *p1, const void *p2)
@@ -120,14 +113,15 @@ format_parse (const char *format, bool translated, char *fdi,
 {
   const char *const format_start = format;
   struct spec spec;
+  unsigned int allocated;
   struct spec *result;
 
   spec.directives = 0;
   spec.named_arg_count = 0;
   spec.unnamed_arg_count = 0;
-  spec.allocated = 0;
   spec.named = NULL;
   spec.unnamed = NULL;
+  allocated = 0;
 
   for (; *format != '\0';)
     if (*format++ == '%')
@@ -191,17 +185,17 @@ format_parse (const char *format, bool translated, char *fdi,
                 goto bad_format;
               }
 
-            if (spec.allocated == spec.unnamed_arg_count)
+            if (allocated == spec.unnamed_arg_count)
               {
-                spec.allocated = 2 * spec.allocated + 1;
-                spec.unnamed = (struct unnamed_arg *) xrealloc (spec.unnamed, spec.allocated * sizeof (struct unnamed_arg));
+                allocated = 2 * allocated + 1;
+                spec.unnamed = (struct unnamed_arg *) xrealloc (spec.unnamed, allocated * sizeof (struct unnamed_arg));
               }
             spec.unnamed[spec.unnamed_arg_count].type = FAT_INTEGER;
             spec.unnamed_arg_count++;
           }
-        else if (isdigit (*format))
+        else if (c_isdigit (*format))
           {
-            do format++; while (isdigit (*format));
+            do format++; while (c_isdigit (*format));
           }
 
         if (*format == '.')
@@ -220,15 +214,15 @@ format_parse (const char *format, bool translated, char *fdi,
                     goto bad_format;
                   }
 
-                if (spec.allocated == spec.unnamed_arg_count)
+                if (allocated == spec.unnamed_arg_count)
                   {
-                    spec.allocated = 2 * spec.allocated + 1;
-                    spec.unnamed = (struct unnamed_arg *) xrealloc (spec.unnamed, spec.allocated * sizeof (struct unnamed_arg));
+                    allocated = 2 * allocated + 1;
+                    spec.unnamed = (struct unnamed_arg *) xrealloc (spec.unnamed, allocated * sizeof (struct unnamed_arg));
                   }
                 spec.unnamed[spec.unnamed_arg_count].type = FAT_INTEGER;
                 spec.unnamed_arg_count++;
               }
-            else if (isdigit (*format))
+            else if (c_isdigit (*format))
               {
                 zero_precision = true;
                 do
@@ -237,7 +231,7 @@ format_parse (const char *format, bool translated, char *fdi,
                       zero_precision = false;
                     format++;
                   }
-                while (isdigit (*format));
+                while (c_isdigit (*format));
               }
           }
 
@@ -288,10 +282,10 @@ format_parse (const char *format, bool translated, char *fdi,
                 goto bad_format;
               }
 
-            if (spec.allocated == spec.named_arg_count)
+            if (allocated == spec.named_arg_count)
               {
-                spec.allocated = 2 * spec.allocated + 1;
-                spec.named = (struct named_arg *) xrealloc (spec.named, spec.allocated * sizeof (struct named_arg));
+                allocated = 2 * allocated + 1;
+                spec.named = (struct named_arg *) xrealloc (spec.named, allocated * sizeof (struct named_arg));
               }
             spec.named[spec.named_arg_count].name = name;
             spec.named[spec.named_arg_count].type = type;
@@ -309,10 +303,10 @@ format_parse (const char *format, bool translated, char *fdi,
                 goto bad_format;
               }
 
-            if (spec.allocated == spec.unnamed_arg_count)
+            if (allocated == spec.unnamed_arg_count)
               {
-                spec.allocated = 2 * spec.allocated + 1;
-                spec.unnamed = (struct unnamed_arg *) xrealloc (spec.unnamed, spec.allocated * sizeof (struct unnamed_arg));
+                allocated = 2 * allocated + 1;
+                spec.unnamed = (struct unnamed_arg *) xrealloc (spec.unnamed, allocated * sizeof (struct unnamed_arg));
               }
             spec.unnamed[spec.unnamed_arg_count].type = type;
             spec.unnamed_arg_count++;
@@ -417,7 +411,7 @@ format_get_number_of_directives (void *descr)
 
 static bool
 format_check (void *msgid_descr, void *msgstr_descr, bool equality,
-              formatstring_error_logger_t error_logger,
+              formatstring_error_logger_t error_logger, void *error_logger_data,
               const char *pretty_msgid, const char *pretty_msgstr)
 {
   struct spec *spec1 = (struct spec *) msgid_descr;
@@ -427,14 +421,16 @@ format_check (void *msgid_descr, void *msgstr_descr, bool equality,
   if (spec1->named_arg_count > 0 && spec2->unnamed_arg_count > 0)
     {
       if (error_logger)
-        error_logger (_("format specifications in '%s' expect a mapping, those in '%s' expect a tuple"),
+        error_logger (error_logger_data,
+                      _("format specifications in '%s' expect a mapping, those in '%s' expect a tuple"),
                       pretty_msgid, pretty_msgstr);
       err = true;
     }
   else if (spec1->unnamed_arg_count > 0 && spec2->named_arg_count > 0)
     {
       if (error_logger)
-        error_logger (_("format specifications in '%s' expect a tuple, those in '%s' expect a mapping"),
+        error_logger (error_logger_data,
+                      _("format specifications in '%s' expect a tuple, those in '%s' expect a mapping"),
                       pretty_msgid, pretty_msgstr);
       err = true;
     }
@@ -446,7 +442,7 @@ format_check (void *msgid_descr, void *msgstr_descr, bool equality,
           unsigned int n1 = spec1->named_arg_count;
           unsigned int n2 = spec2->named_arg_count;
 
-          /* Check the argument names are the same.
+          /* Check the argument names in spec2 are contained in those of spec1.
              Both arrays are sorted.  We search for the first difference.  */
           for (i = 0, j = 0; i < n1 || j < n2; )
             {
@@ -457,7 +453,8 @@ format_check (void *msgid_descr, void *msgstr_descr, bool equality,
               if (cmp > 0)
                 {
                   if (error_logger)
-                    error_logger (_("a format specification for argument '%s', as in '%s', doesn't exist in '%s'"),
+                    error_logger (error_logger_data,
+                                  _("a format specification for argument '%s', as in '%s', doesn't exist in '%s'"),
                                   spec2->named[j].name, pretty_msgstr,
                                   pretty_msgid);
                   err = true;
@@ -468,7 +465,8 @@ format_check (void *msgid_descr, void *msgstr_descr, bool equality,
                   if (equality)
                     {
                       if (error_logger)
-                        error_logger (_("a format specification for argument '%s' doesn't exist in '%s'"),
+                        error_logger (error_logger_data,
+                                      _("a format specification for argument '%s' doesn't exist in '%s'"),
                                       spec1->named[i].name, pretty_msgstr);
                       err = true;
                       break;
@@ -491,7 +489,8 @@ format_check (void *msgid_descr, void *msgstr_descr, bool equality,
                                   || spec2->named[j].type == FAT_ANY))))
                       {
                         if (error_logger)
-                          error_logger (_("format specifications in '%s' and '%s' for argument '%s' are not the same"),
+                          error_logger (error_logger_data,
+                                        _("format specifications in '%s' and '%s' for argument '%s' are not the same"),
                                         pretty_msgid, pretty_msgstr,
                                         spec2->named[j].name);
                         err = true;
@@ -512,7 +511,8 @@ format_check (void *msgid_descr, void *msgstr_descr, bool equality,
           if (spec1->unnamed_arg_count != spec2->unnamed_arg_count)
             {
               if (error_logger)
-                error_logger (_("number of format specifications in '%s' and '%s' does not match"),
+                error_logger (error_logger_data,
+                              _("number of format specifications in '%s' and '%s' does not match"),
                               pretty_msgid, pretty_msgstr);
               err = true;
             }
@@ -524,7 +524,8 @@ format_check (void *msgid_descr, void *msgstr_descr, bool equality,
                             || spec2->unnamed[i].type == FAT_ANY))))
                 {
                   if (error_logger)
-                    error_logger (_("format specifications in '%s' and '%s' for argument %u are not the same"),
+                    error_logger (error_logger_data,
+                                  _("format specifications in '%s' and '%s' for argument %u are not the same"),
                                   pretty_msgid, pretty_msgstr, i + 1);
                   err = true;
                 }
@@ -688,7 +689,7 @@ main ()
 /*
  * For Emacs M-x compile
  * Local Variables:
- * compile-command: "/bin/sh ../libtool --tag=CC --mode=link gcc -o a.out -static -O -g -Wall -I.. -I../gnulib-lib -I../intl -DHAVE_CONFIG_H -DTEST format-python.c ../gnulib-lib/libgettextlib.la"
+ * compile-command: "/bin/sh ../libtool --tag=CC --mode=link gcc -o a.out -static -O -g -Wall -I.. -I../gnulib-lib -I../../gettext-runtime/intl -DHAVE_CONFIG_H -DTEST format-python.c ../gnulib-lib/libgettextlib.la"
  * End:
  */
 
