@@ -1,5 +1,5 @@
 /* Message list charset and locale charset handling.
-   Copyright (C) 2001-2003, 2005-2009, 2015-2016 Free Software Foundation, Inc.
+   Copyright (C) 2001-2025 Free Software Foundation, Inc.
    Written by Bruno Haible <haible@clisp.cons.org>, 2001.
 
    This program is free software: you can redistribute it and/or modify
@@ -13,7 +13,7 @@
    GNU General Public License for more details.
 
    You should have received a copy of the GNU General Public License
-   along with this program.  If not, see <http://www.gnu.org/licenses/>.  */
+   along with this program.  If not, see <https://www.gnu.org/licenses/>.  */
 
 
 #ifdef HAVE_CONFIG_H
@@ -32,18 +32,22 @@
 # include <iconv.h>
 #endif
 
+#include "noreturn.h"
 #include "progname.h"
-#include "basename.h"
+#include "basename-lgpl.h"
+#include "string-desc.h"
 #include "message.h"
 #include "po-charset.h"
+#include "msgl-header.h"
 #include "xstriconv.h"
 #include "xstriconveh.h"
 #include "msgl-ascii.h"
+#include "msgl-ofn.h"
 #include "xalloc.h"
 #include "xmalloca.h"
 #include "c-strstr.h"
 #include "xvasprintf.h"
-#include "po-xerror.h"
+#include "xerror-handler.h"
 #include "gettext.h"
 
 #define _(str) gettext (str)
@@ -51,25 +55,23 @@
 
 #if HAVE_ICONV
 
-static void conversion_error (const struct conversion_context* context)
-#if defined __GNUC__ && ((__GNUC__ == 2 && __GNUC_MINOR__ >= 5) || __GNUC__ > 2)
-     __attribute__ ((noreturn))
-#endif
-;
+_GL_NORETURN_FUNC static void
+conversion_error (const struct conversion_context* context,
+                  xerror_handler_ty xeh);
 static void
-conversion_error (const struct conversion_context* context)
+conversion_error (const struct conversion_context* context,
+                  xerror_handler_ty xeh)
 {
   if (context->to_code == po_charset_utf8)
     /* If a conversion to UTF-8 fails, the problem lies in the input.  */
-    po_xerror (PO_SEVERITY_FATAL_ERROR, context->message, NULL, 0, 0, false,
-               xasprintf (_("%s: input is not valid in \"%s\" encoding"),
-                          context->from_filename, context->from_code));
+    xeh->xerror (CAT_SEVERITY_FATAL_ERROR, context->message, NULL, 0, 0, false,
+                 xasprintf (_("%s: input is not valid in \"%s\" encoding"),
+                            context->from_filename, context->from_code));
   else
-    po_xerror (PO_SEVERITY_FATAL_ERROR, context->message, NULL, 0, 0, false,
-               xasprintf (_("\
-%s: error while converting from \"%s\" encoding to \"%s\" encoding"),
-                          context->from_filename, context->from_code,
-                          context->to_code));
+    xeh->xerror (CAT_SEVERITY_FATAL_ERROR, context->message, NULL, 0, 0, false,
+                 xasprintf (_("%s: error while converting from \"%s\" encoding to \"%s\" encoding"),
+                            context->from_filename, context->from_code,
+                            context->to_code));
   /* NOTREACHED */
   abort ();
 }
@@ -88,14 +90,31 @@ convert_string_directly (iconv_t cd, const char *string,
         && strlen (result) == resultlen - 1)
       return result;
 
-  conversion_error (context);
+  conversion_error (context, textmode_xerror_handler);
   /* NOTREACHED */
   return NULL;
 }
 
+string_desc_t
+convert_string_desc_directly (iconv_t cd, string_desc_t string,
+                              const struct conversion_context* context)
+{
+  char *result = NULL;
+  size_t resultlen = 0;
+
+  if (xmem_cd_iconv (sd_data (string), sd_length (string),
+                     cd, &result, &resultlen) == 0)
+    return sd_new_addr (resultlen, result);
+
+  conversion_error (context, textmode_xerror_handler);
+  /* NOTREACHED */
+  return sd_new_empty ();
+}
+
 static char *
 convert_string (const iconveh_t *cd, const char *string,
-                const struct conversion_context* context)
+                const struct conversion_context* context,
+                xerror_handler_ty xeh)
 {
   size_t len = strlen (string) + 1;
   char *result = NULL;
@@ -108,48 +127,53 @@ convert_string (const iconveh_t *cd, const char *string,
         && strlen (result) == resultlen - 1)
       return result;
 
-  conversion_error (context);
+  conversion_error (context, xeh);
   /* NOTREACHED */
   return NULL;
 }
 
 static void
 convert_string_list (const iconveh_t *cd, string_list_ty *slp,
-                     const struct conversion_context* context)
+                     const struct conversion_context* context,
+                     xerror_handler_ty xeh)
 {
   size_t i;
 
   if (slp != NULL)
     for (i = 0; i < slp->nitems; i++)
-      slp->item[i] = convert_string (cd, slp->item[i], context);
+      slp->item[i] = convert_string (cd, slp->item[i], context, xeh);
 }
 
 static void
 convert_prev_msgid (const iconveh_t *cd, message_ty *mp,
-                    const struct conversion_context* context)
+                    const struct conversion_context* context,
+                    xerror_handler_ty xeh)
 {
   if (mp->prev_msgctxt != NULL)
-    mp->prev_msgctxt = convert_string (cd, mp->prev_msgctxt, context);
+    mp->prev_msgctxt = convert_string (cd, mp->prev_msgctxt, context, xeh);
   if (mp->prev_msgid != NULL)
-    mp->prev_msgid = convert_string (cd, mp->prev_msgid, context);
+    mp->prev_msgid = convert_string (cd, mp->prev_msgid, context, xeh);
   if (mp->prev_msgid_plural != NULL)
-    mp->prev_msgid_plural = convert_string (cd, mp->prev_msgid_plural, context);
+    mp->prev_msgid_plural =
+      convert_string (cd, mp->prev_msgid_plural, context, xeh);
 }
 
 static void
 convert_msgid (const iconveh_t *cd, message_ty *mp,
-               const struct conversion_context* context)
+               const struct conversion_context* context,
+               xerror_handler_ty xeh)
 {
   if (mp->msgctxt != NULL)
-    mp->msgctxt = convert_string (cd, mp->msgctxt, context);
-  mp->msgid = convert_string (cd, mp->msgid, context);
+    mp->msgctxt = convert_string (cd, mp->msgctxt, context, xeh);
+  mp->msgid = convert_string (cd, mp->msgid, context, xeh);
   if (mp->msgid_plural != NULL)
-    mp->msgid_plural = convert_string (cd, mp->msgid_plural, context);
+    mp->msgid_plural = convert_string (cd, mp->msgid_plural, context, xeh);
 }
 
 static void
 convert_msgstr (const iconveh_t *cd, message_ty *mp,
-                const struct conversion_context* context)
+                const struct conversion_context* context,
+                xerror_handler_ty xeh)
 {
   char *result = NULL;
   size_t resultlen = 0;
@@ -183,7 +207,7 @@ convert_msgstr (const iconveh_t *cd, message_ty *mp,
           }
       }
 
-  conversion_error (context);
+  conversion_error (context, xeh);
 }
 
 #endif
@@ -194,7 +218,8 @@ iconv_message_list_internal (message_list_ty *mlp,
                              const char *canon_from_code,
                              const char *canon_to_code,
                              bool update_header,
-                             const char *from_filename)
+                             const char *from_filename,
+                             xerror_handler_ty xeh)
 {
   bool canon_from_code_overridden = (canon_from_code != NULL);
   bool msgids_changed;
@@ -231,22 +256,25 @@ iconv_message_list_internal (message_list_ty *mlp,
                   {
                     if (!canon_from_code_overridden)
                       {
-                        /* Don't give an error for POT files, because POT
-                           files usually contain only ASCII msgids.  */
+                        /* Don't give an error for POT files, because
+                           POT files usually contain only ASCII msgids.
+                           Also don't give an error for disguised POT
+                           files that actually contain only ASCII msgids.  */
                         const char *filename = from_filename;
                         size_t filenamelen;
 
-                        if (filename != NULL
-                            && (filenamelen = strlen (filename)) >= 4
-                            && memcmp (filename + filenamelen - 4, ".pot", 4)
-                               == 0
-                            && strcmp (charset, "CHARSET") == 0)
+                        if (strcmp (charset, "CHARSET") == 0
+                            && ((filename != NULL
+                                 && (filenamelen = strlen (filename)) >= 4
+                                 && memcmp (filename + filenamelen - 4, ".pot", 4)
+                                    == 0)
+                                || is_ascii_message_list (mlp)))
                           canon_charset = po_charset_ascii;
                         else
-                          po_xerror (PO_SEVERITY_FATAL_ERROR, NULL, NULL, 0, 0,
-                                     false, xasprintf (_("\
-present charset \"%s\" is not a portable encoding name"),
-                                                charset));
+                          xeh->xerror (CAT_SEVERITY_FATAL_ERROR, NULL, NULL, 0, 0,
+                                       false,
+                                       xasprintf (_("present charset \"%s\" is not a portable encoding name"),
+                                                  charset));
                       }
                   }
                 else
@@ -254,30 +282,15 @@ present charset \"%s\" is not a portable encoding name"),
                     if (canon_from_code == NULL)
                       canon_from_code = canon_charset;
                     else if (canon_from_code != canon_charset)
-                      po_xerror (PO_SEVERITY_FATAL_ERROR, NULL, NULL, 0,  0,
-                                 false,
-                                 xasprintf (_("\
-two different charsets \"%s\" and \"%s\" in input file"),
-                                            canon_from_code, canon_charset));
+                      xeh->xerror (CAT_SEVERITY_FATAL_ERROR, NULL, NULL, 0,  0,
+                                   false,
+                                   xasprintf (_("two different charsets \"%s\" and \"%s\" in input file"),
+                                              canon_from_code, canon_charset));
                   }
                 freea (charset);
 
                 if (update_header)
-                  {
-                    size_t len1, len2, len3;
-                    char *new_header;
-
-                    len1 = charsetstr - header;
-                    len2 = strlen (canon_to_code);
-                    len3 = (header + strlen (header)) - (charsetstr + len);
-                    new_header = XNMALLOC (len1 + len2 + len3 + 1, char);
-                    memcpy (new_header, header, len1);
-                    memcpy (new_header + len1, canon_to_code, len2);
-                    memcpy (new_header + len1 + len2, charsetstr + len,
-                            len3 + 1);
-                    mlp->item[j]->msgstr = new_header;
-                    mlp->item[j]->msgstr_len = len1 + len2 + len3 + 1;
-                  }
+                  header_set_charset (mlp->item[j], charsetstr, canon_to_code);
               }
           }
       }
@@ -286,9 +299,8 @@ two different charsets \"%s\" and \"%s\" in input file"),
       if (is_ascii_message_list (mlp))
         canon_from_code = po_charset_ascii;
       else
-        po_xerror (PO_SEVERITY_FATAL_ERROR, NULL, NULL, 0, 0, false,
-                   _("\
-input file doesn't contain a header entry with a charset specification"));
+        xeh->xerror (CAT_SEVERITY_FATAL_ERROR, NULL, NULL, 0, 0, false,
+                     _("input file doesn't contain a header entry with a charset specification"));
     }
 
   msgids_changed = false;
@@ -301,12 +313,10 @@ input file doesn't contain a header entry with a charset specification"));
       struct conversion_context context;
 
       if (iconveh_open (canon_to_code, canon_from_code, &cd) < 0)
-        po_xerror (PO_SEVERITY_FATAL_ERROR, NULL, NULL, 0, 0, false,
-                   xasprintf (_("\
-Cannot convert from \"%s\" to \"%s\". %s relies on iconv(), \
-and iconv() does not support this conversion."),
-                              canon_from_code, canon_to_code,
-                              basename (program_name)));
+        xeh->xerror (CAT_SEVERITY_FATAL_ERROR, NULL, NULL, 0, 0, false,
+                     xasprintf (_("Cannot convert from \"%s\" to \"%s\". %s relies on iconv(), and iconv() does not support this conversion."),
+                                canon_from_code, canon_to_code,
+                                last_component (program_name)));
 
       context.from_code = canon_from_code;
       context.to_code = canon_to_code;
@@ -320,29 +330,25 @@ and iconv() does not support this conversion."),
               || !is_ascii_string (mp->msgid))
             msgids_changed = true;
           context.message = mp;
-          convert_string_list (&cd, mp->comment, &context);
-          convert_string_list (&cd, mp->comment_dot, &context);
-          convert_prev_msgid (&cd, mp, &context);
-          convert_msgid (&cd, mp, &context);
-          convert_msgstr (&cd, mp, &context);
+          convert_string_list (&cd, mp->comment, &context, xeh);
+          convert_string_list (&cd, mp->comment_dot, &context, xeh);
+          convert_prev_msgid (&cd, mp, &context, xeh);
+          convert_msgid (&cd, mp, &context, xeh);
+          convert_msgstr (&cd, mp, &context, xeh);
         }
 
       iconveh_close (&cd);
 
       if (msgids_changed)
         if (message_list_msgids_changed (mlp))
-          po_xerror (PO_SEVERITY_FATAL_ERROR, NULL, NULL, 0, 0, false,
-                     xasprintf (_("\
-Conversion from \"%s\" to \"%s\" introduces duplicates: \
-some different msgids become equal."),
-                                canon_from_code, canon_to_code));
+          xeh->xerror (CAT_SEVERITY_FATAL_ERROR, NULL, NULL, 0, 0, false,
+                       xasprintf (_("Conversion from \"%s\" to \"%s\" introduces duplicates: some different msgids become equal."),
+                                  canon_from_code, canon_to_code));
 #else
-          po_xerror (PO_SEVERITY_FATAL_ERROR, NULL, NULL, 0, 0, false,
-                     xasprintf (_("\
-Cannot convert from \"%s\" to \"%s\". %s relies on iconv(). \
-This version was built without iconv()."),
-                                canon_from_code, canon_to_code,
-                                basename (program_name)));
+          xeh->xerror (CAT_SEVERITY_FATAL_ERROR, NULL, NULL, 0, 0, false,
+                       xasprintf (_("Cannot convert from \"%s\" to \"%s\". %s relies on iconv(). This version was built without iconv()."),
+                                  canon_from_code, canon_to_code,
+                                  last_component (program_name)));
 #endif
     }
 
@@ -352,18 +358,19 @@ This version was built without iconv()."),
 bool
 iconv_message_list (message_list_ty *mlp,
                     const char *canon_from_code, const char *canon_to_code,
-                    const char *from_filename)
+                    const char *from_filename, xerror_handler_ty xeh)
 {
   return iconv_message_list_internal (mlp,
                                       canon_from_code, canon_to_code, true,
-                                      from_filename);
+                                      from_filename, xeh);
 }
 
 msgdomain_list_ty *
 iconv_msgdomain_list (msgdomain_list_ty *mdlp,
                       const char *to_code,
                       bool update_header,
-                      const char *from_filename)
+                      const char *from_filename,
+                      xerror_handler_ty xeh)
 {
   const char *canon_to_code;
   size_t k;
@@ -371,15 +378,23 @@ iconv_msgdomain_list (msgdomain_list_ty *mdlp,
   /* Canonicalize target encoding.  */
   canon_to_code = po_charset_canonicalize (to_code);
   if (canon_to_code == NULL)
-    po_xerror (PO_SEVERITY_FATAL_ERROR, NULL, NULL, 0, 0, false,
-               xasprintf (_("\
-target charset \"%s\" is not a portable encoding name."),
-                          to_code));
+    xeh->xerror (CAT_SEVERITY_FATAL_ERROR, NULL, NULL, 0, 0, false,
+                 xasprintf (_("target charset \"%s\" is not a portable encoding name."),
+                            to_code));
+
+  /* Test whether the control characters required for escaping file names with
+     spaces are present in the target encoding.  */
+  if (msgdomain_list_has_filenames_with_spaces (mdlp)
+      && !(canon_to_code == po_charset_utf8
+           || strcmp (canon_to_code, "GB18030") == 0))
+    xeh->xerror (CAT_SEVERITY_FATAL_ERROR, NULL, NULL, 0, 0, false,
+                 xasprintf (_("Cannot write the control characters that protect file names with spaces in the %s encoding"),
+                            canon_to_code));
 
   for (k = 0; k < mdlp->nitems; k++)
     iconv_message_list_internal (mdlp->item[k]->messages,
                                  mdlp->encoding, canon_to_code, update_header,
-                                 from_filename);
+                                 from_filename, xeh);
 
   mdlp->encoding = canon_to_code;
   return mdlp;
